import torch
import numpy  as np

def psnr(output, label):
    '''
    :param output:
    :param label:
    :return: PSNR value calculated in pytorch way
    '''
    mse = torch.mean((output - label) ** 2)
    return 20 * torch.log10(1 / torch.sqrt(mse))

def psnr_np(output, label):
    '''
    :param output:
    :param label:
    :return: PSNR value calculated in pytorch way
    '''
    mse = np.mean((output - label) ** 2)
    return 20 * np.log10(1 / np.sqrt(mse))


from networks import MSDNet, UNet
from .dataset import CustomImageDataset, CustomGreyImageDataset
from .metric import psnr, psnr_np
from .train import train_loop, test_loop, valid_loop
from .utils import count_parameters


import os
import glob
import numpy as np
from torch.utils.data import Dataset
from PIL import Image


class CustomImageDataset(Dataset):
    def __init__(self, img_dir, width, height, vmax=255, transform=None, target_transform=None):
        '''
        :param img_dir: The directory where all images are located
        :param width: The desired height of output image's width
        :param height: The desired height of output image's height
        :param vmax: The max possible value of pixel of given image, used to normalize the images
        :param transform: The transformation function for images
        :param target_transform: The transformation function for images targets(labels)
        '''
        self.paths = glob.glob(os.path.join(img_dir, "*.JPEG"))
        self.width = width
        self.height = height
        self.vmax = vmax
        self.transform = transform
        self.target_transform = target_transform

    def __len__(self):
        return len(self.paths)

    def __getitem__(self, idx):
        image = Image.open(self.paths[idx])
        image = image.resize((self.width, self.height), Image.ANTIALIAS)
        if (image.mode != 'RGB'):
            image = image.convert("RGB")
        image = np.asarray(image)/self.vmax
        label = image.copy()
        image = image.transpose(2, 0, 1)
        label = label.transpose(2, 0, 1)

        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)
        return image, label

class CustomGreyImageDataset(Dataset):
    def __init__(self, img_dir, width, height, vmax=255, transform=None, target_transform=None):
        '''
        :param img_dir: The directory where all images are located
        :param width: The desired height of output image's width
        :param height: The desired height of output image's height
        :param vmax: The max possible value of pixel of given image, used to normalize the images
        :param transform: The transformation function for images
        :param target_transform: The transformation function for images targets(labels)
        '''
        self.paths = glob.glob(os.path.join(img_dir, "*.JPEG"))
        self.width = width
        self.height = height
        self.vmax = vmax
        self.transform = transform
        self.target_transform = target_transform

    def __len__(self):
        return len(self.paths)

    def __getitem__(self, idx):
        image = Image.open(self.paths[idx])
        image = image.resize((self.width, self.height), Image.ANTIALIAS)
        if (image.mode == 'RGB'):
            image = image.convert("L")
        image = np.asarray(image)/self.vmax
        label = image.copy()
        image = np.expand_dims(image, (0))
        label = np.expand_dims(label, (0))

        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)
        return image, label


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(python_requires='>=3.7',
    name='pcfv',
    version='0.0.4',
    description='The common used function with pytorch for vision tasks',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
    'Development Status :: 1 - Planning',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    url='https://github.com/jiayangshi/pcf',
    author='Jiayang Shi',
    author_email='j.shi@liacs.leidenuniv.nl',
    license='MIT',
    packages=['pcfv'],
    install_requires=[
        'pytorch',
        'numpy',
        'matplotlib',
        'Pillow'
    ],
    zip_safe=False)



import os
import torch
import numpy as np
import matplotlib.pyplot as plt

def train_loop(dataloader, model, optimizer, loss, device='cuda'):
    '''
    The training function
    :param dataloader: The dataloader to provide the data
    :param model: The to be trained model
    :param optimizer: The optimizer
    :param loss: The used loss function
    :param device: The device on which the training is done
    :return: The average training loss
    '''
    training_loss = 0
    size = len(dataloader.dataset)
    batches = len(dataloader)
    for batch, (X, y) in enumerate(dataloader):
        # Compute prediction and loss
        X = X.to(device, dtype=torch.float)
        pred = model(X)
        label = y.to(device, dtype=torch.float)
        loss = loss(pred, label)

        # Backpropagation
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # display current batch and loss
        loss, current = loss.item(), batch * len(X)
        print(f"training loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")

        # calculates the average training loss
        training_loss += loss/batches

    return training_loss

def valid_loop(dataloader, model, loss, device='cuda'):
    '''
    The validation function
    :param dataloader: The dataloader to provide the data
    :param model: The to be trained model
    :param loss: The used loss function
    :param device: The device on which the validation is done
    :return: The average validation loss
    '''
    valid_loss = 0
    size = len(dataloader.dataset)
    batches = len(dataloader)
    for batch, (X, y) in enumerate(dataloader):
        # Compute prediction and loss
        X = X.to(device, dtype=torch.float)
        pred = model(X)
        label = y.to(device, dtype=torch.float)
        loss = loss(pred, label)

        # display current batch and loss
        loss, current = loss.item(), batch * len(X)
        print(f"validation loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")

        # calculates the average training loss
        valid_loss += loss/batches
    return valid_loss

def test_loop(dataloader, model, loss, metric, output_directory=None,  device='cuda'):
    '''

    :param dataloader: The dataloader to provide the data
    :param model: The to be trained model
    :param loss: The used loss function
    :param metric: The used metric function
    :param output_directory: The directory to save the test results
    :param device: The device on which the validation is done
    '''
    batches = len(dataloader)
    test_loss, test_metric = 0, 0

    i = 0
    with torch.no_grad():
        for X, y in dataloader:
            X = X.to(device, dtype=torch.float)
            pred = model(X)
            label = y.to(device, dtype=torch.float)
            loss = loss(pred, label)
            metric = metric(pred, label)
            if not output_directory:
                for j in range(pred.shape[0]):
                    fig = plt.figure(frameon=True)
                    ax1 = plt.subplot(1, 3, 1)
                    ax1.imshow(np.squeeze(label[j].cpu().numpy()), vmin=0, vmax=1)
                    plt.xticks([])
                    plt.yticks([])
                    ax1.set_title("Original")
                    ax2 = plt.subplot(1, 3, 2)
                    ax2.imshow(np.squeeze(X[j].cpu().numpy()), vmin=0, vmax=1)
                    plt.xticks([])
                    plt.yticks([])
                    ax2.set_title("Noised")
                    ax2.set_xlabel("PSNR:{:,.2f} dB".format(metric(label[j], X[j]).cpu().numpy()))
                    ax3 = plt.subplot(1, 3, 3)
                    ax3.imshow(np.squeeze(pred[j].cpu().numpy()), vmin=0, vmax=1)
                    plt.xticks([])
                    plt.yticks([])
                    ax3.set_title("Denoised")
                    ax3.set_xlabel("PSNR:{:,.2f} dB".format(metric(label[j], X[j]).cpu().numpy()))
                    fig.savefig(os.path.join(output_directory, str(i) + ".png"))
                    print("The {}th test image is processed".format(i + 1))
                    i += 1
            test_loss += loss / batches
            test_metric += metric / batches

    print(f"Avg loss on whole image: {test_loss:>8f} \n")
    print(f"Avg metric on whole image: {test_metric:>8f} \n")